#include "ImageManager.h"

int ImageManager::imgBufferPush(string imageName, string imageFormat) {
	localBuffer = imageEncoder(imageName, imageFormat);
	for (int i = localBuffer.size() - 1; i >=0 ; i--) {
		localBufferStack.push( localBuffer[i] );
	}

	return localBuffer.size();
}

vector<unsigned char> ImageManager::imgBufferPop(int imgPktSize) {
	vector<unsigned char> auxLocalBuffer;
	int i = 0;
	while( i < imgPktSize && !localBufferStack.empty() ){
		auxLocalBuffer.push_back( localBufferStack.top() );
		localBufferStack.pop();
		i++;
	}

	visualCastaliaTrace.writeTrace( numNode, simulationTime,
					 "Get a fragment of buffer of "+visualCastaliaTrace.intToString(auxLocalBuffer.size())+ " bytes");
	return auxLocalBuffer;
}

void ImageManager::storeRcvImage (vector<unsigned char> buffer, string path_to_save, int source) {
	imageDecoder(buffer, path_to_save, source);
}

//						==============================
// 						FUNCTIONS RELATED WITH OPENCV
//						==============================

// This function encode an image and returns the generated buffer
vector<unsigned char> ImageManager::imageEncoder(string imageName, string imageFormat){
	vector<unsigned char> buffer;

	// Windows bitmaps - *.bmp, *.dib (always supported)
	// JPEG files - *.jpeg, *.jpg, *.jpe (see the Notes section)
	// JPEG 2000 files - *.jp2 (see the Notes section)
	// Portable Network Graphics - *.png (see the Notes section)
	Mat image = imread(imageName.c_str(), IMREAD_UNCHANGED);
	
	// Check for invalid input
	if( image.empty() ) {
        string infoToTrace;
        infoToTrace =  "Could not open or find the image files at simulation directory, path: " + imageName;
        visualCastaliaTrace.writeTrace( numNode, simulationTime,
					 infoToTrace);
    }
	else {
		//namedWindow( "Display window", WINDOW_AUTOSIZE ); // Create a window for display.
    	//imshow( "Display window", image );                // Show our image inside it.
    	//waitKey(1); // Wait for a keystroke in the window
		
		int compressionMode;		
    	if(imageFormat == ".jpeg" || imageFormat == ".jpg" || imageFormat == ".jpe" || imageFormat == ".jp2")
    		compressionMode = JPG_MODE;
    	else if (imageFormat == ".png")
    		compressionMode = PNG_MODE;
    	else if (imageFormat == ".bmp" || imageFormat == ".dib")
    		compressionMode = BITMAP_MODE;

    	vector<int> compression_param = vector<int>(2);
		int flag;		
		int compressionLevel;
		string imageFormat;

		switch(compressionMode){
			case JPG_MODE: {
				// Max quality of compression - almost no compression
				flag = IMWRITE_JPEG_QUALITY;
				compressionLevel = 100;
				imageFormat = ".jpg";
				break;
			}
			case BITMAP_MODE: {
				imageFormat = ".bmp";
				break;
			}
			case PNG_MODE: {
				// For PNG, it can be the compression level from 0 to 9. A higher value means a 
				// smaller size and longer compression time. Default value is 3.
				imageFormat = ".png";
				flag = IMWRITE_PNG_COMPRESSION;
				compressionLevel = 0;
				break;
			}
			default: {

			}
		}
		compression_param[0] = flag;
		compression_param[1] = compressionLevel;

		// The function imencode compresses the image and stores it in the memory buffer
		// that is resized to fit the result. See cv::imwrite for the list of supported 
		// formats and flags description.

		imencode(imageFormat.c_str(), image, buffer, compression_param);
		visualCastaliaTrace.writeTrace( numNode, simulationTime,
					 "Buffer to image: " +
					  imageName +
					   " created - size: "+
					   visualCastaliaTrace.intToString( buffer.size() ) );
	}
		return buffer;
}

void ImageManager::imageDecoder(vector<unsigned char> buffer, string path_to_save, int source){
	// Decode buffer into an image
	// The function reads an image from the specified buffer in the memory.
	// If the buffer is too short or contains invalid data, the empty matrix/image is returned.
	Mat decodedImage = imdecode( Mat(buffer), IMREAD_UNCHANGED );

	if( decodedImage.empty() ) {
		visualCastaliaTrace.writeTrace( numNode, simulationTime,
					 "Fail in decoding process of buffer from image: "+path_to_save);
	}
	else{
		// This function saves the decoded image
		bool isSaved = imwrite( path_to_save.c_str(), decodedImage );
		
		if (isSaved){
			// Writing trace to indicate when the image is saved
			visualCastaliaTrace.writeTrace( numNode, simulationTime,
									 "Saving image from source node #"+
									 visualCastaliaTrace.intToString(source)+
									 " in: "+path_to_save);
		}
		else{
			visualCastaliaTrace.writeTrace( numNode, simulationTime,
						 "Received image is corrupted and cannot be saved in: "+path_to_save);
		}
	}

}

void ImageManager::generateRandomImages(int imgHeight, int imgWidth, bool useColoredImages, 
										 string path_to_save, int caseToDraw){
	// RGB codification to set different colors to each image case
	int Red, Blue, Green;
	// Creating an empty image frame
	Mat image(imgHeight, imgWidth, CV_8UC3);
	// Creating the opposite points to position the inner rectangle,
	// according the dimensions of image
	Point pt1, pt2;
	pt1.x = -imgWidth/2;
	pt1.y = -imgHeight/2;
	pt2.x = imgWidth/2;
	pt2.y = imgHeight/2;
	// Type of line of the geometry
	int lineType = 8;
	// when negative, it indicates that the geometry is completely filled
	int thickness = -1;
    // The function rand() generates numbers between 0 and 255 in this case
    Blue = rand() % 255;
    Green = rand() % 255;
    Red = rand() % 255;
	// Coloring the image according RGB codification
	image = Scalar(Blue, Green, Red);
	// Creating a rectangle inside the image
	rectangle( image, pt2, pt1, Scalar::all(0), thickness, lineType );
	// Converting the case number to string to put it like inner text in the generated image
	ostringstream convert;	
	convert << caseToDraw;
	// Drawing the number of case in the generated image
	thickness = 3;
	putText(image, convert.str(), pt2, FONT_HERSHEY_SIMPLEX, 4, Scalar::all(0), thickness, 8, false);
	// Setting the png parameters to save the image in this format
    vector<int> compression_params;
    // Type of compression
    compression_params.push_back(IMWRITE_JPEG_QUALITY);
    // Compression level - 3: default value
    compression_params.push_back(100);

    // The image is saved in grey scale if the useColoredImages is false
    if (!useColoredImages){
    	Mat gray_image;
    	// Converting the generated image to grey scale
    	cvtColor( image, gray_image, COLOR_BGR2GRAY );
    	imwrite( path_to_save, gray_image, compression_params );
    }
    else{
		imwrite( path_to_save, image, compression_params );	
    }
    visualCastaliaTrace.writeTrace( numNode, simulationTime,
					 "Image: " + path_to_save + " generated");
}
